/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlImageLoader.cpp
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**      based on applications by Stephen Gould and Sid Batra
**             
** DESCRIPTION:
**   Commonly re-used subroutines for processing command line arguments related
**   to patch types, mapping from channel types to patch types, or reading
**   and writing image files in the patch based object detector pipeline
**
*****************************************************************************/
#include "svlImageLoader.h"
#include "svlImageSequence.h"

// If true, then frames are associated with a mask matrix file, where
// 1s indicate that pixels should be treated as unobserved. Right now
// the only use of this is to block buildPatchDictionary from using
// certain image regions
bool svlImageLoader::USE_MASK = false;
string svlImageLoader::MASK_EXT = string(".occ");
string svlImageLoader::BASE_EXT = string(".jpg");
int svlImageLoader::DEPTH_MAP_WIDTH = -1;
int svlImageLoader::DEPTH_MAP_HEIGHT = -1;
int svlImageLoader::WIDTH = -1;
int svlImageLoader::HEIGHT = -1;
vector<pair<string, string> > svlImageLoader::CHANNELS;
vector<string> svlImageLoader::EXT_TO_IGNORE;

// combination of parseCommandLineItem and finishCommandLine from before
svlImageLoader::svlImageLoader(bool default_use_edge) : 
  _useMask(USE_MASK), _maskExt(MASK_EXT),
  _depthMapWidth(DEPTH_MAP_WIDTH),
  _depthMapHeight(DEPTH_MAP_HEIGHT),
  _width(WIDTH), _height(HEIGHT)
{
  if (_depthMapWidth == -1 || _depthMapHeight == -1) {
    _depthMapWidth = WIDTH;
    _depthMapHeight = HEIGHT;
  }

  // Parsing the channels
  if (CHANNELS.size() == 0) { // create the default channels
    
    //Always assume there is an INTENSITY channel.
    _channelTypes.push_back(SVL_CHANNEL_INTENSITY);
    _colorTypes.push_back(SVL_COLOR_GRAY);
    _channelPatchTypes.push_back(SVL_INTENSITY_PATCH);
    _channelExts.push_back(BASE_EXT);
    
    // If necessary, populate the edge channel as well
    if (default_use_edge) {
      _channelTypes.push_back(SVL_CHANNEL_EDGE);
      _colorTypes.push_back(SVL_COLOR_GRAY);
      _channelPatchTypes.push_back(SVL_INTENSITY_PATCH);
      _channelExts.push_back(BASE_EXT);
    }
  } else { // channels specified
    for (unsigned i = 0; i < CHANNELS.size(); i++) {
      const char *type = CHANNELS[i].first.c_str();
      const char *ext = CHANNELS[i].second.c_str();
      
      svlChannelType chType = textToChannelType(type);

      SVL_ASSERT_MSG(chType != SVL_CHANNEL_ERROR, 
		     "unknown channel type " << type);

      svlColorType crType = textToColorType(type);

      SVL_ASSERT_MSG(chType != SVL_CHANNEL_ERROR, 
		     "unknown color type " << type);
      
      svlPatchDefinitionType patchType = channelTypeToPatchType(chType);
      
      _channelTypes.push_back(chType);
      _colorTypes.push_back(crType);
      _channelPatchTypes.push_back(patchType);
      
      _channelExts.push_back((!strcmp(ext, "-")) ? BASE_EXT : ext);
    }
  }

  //use this to tell eTree not to load files ending in some
  //extension. TODO-- make patch dictionary smart enough to ignore
  //arbitrary channels, right now you can only ignore channels at the
  //end or the patch dictionary will come out with the wrong
  //validChannel indices
  for (unsigned i = 0; i < EXT_TO_IGNORE.size(); i++) {
    _eTree.ignore(EXT_TO_IGNORE[i].c_str());
  }

  // Prepare for loading files later on
  for (unsigned i = 0; i < _channelExts.size(); i++) {
    _files.push_back(vector<string> ());
    _eTree.addChannel(i, _channelExts[i]);
  }
  
  if (_useMask) {
    _files.push_back(vector<string> ());
    _eTree.addChannel(_channelExts.size(), _maskExt);
  }

}

svlImageLoader::svlImageLoader(const svlImageLoader & o) :
   _channelTypes(o._channelTypes), _colorTypes(o._colorTypes), 
   _channelPatchTypes(o._channelPatchTypes), _channelExts(o._channelExts), 
   _softEdgeCalculator(o._softEdgeCalculator), 
   _useMask(o._useMask), _maskExt(o._maskExt),
   _dir(o._dir), _eTree(o._eTree), _files(o._files),
   _depthMapWidth(o._depthMapWidth), _depthMapHeight(o._depthMapHeight),
   _width(o._width), _height(o._height)
{
  // do nothing
}

static string stripExtension(const string & str, const string & ext)
{
  assert(str.size() >= ext.size());
  return str.substr(0, str.size()-ext.size());
}

void svlImageLoader::readDir(const char * dir)
{
  _dir = dir;

  _eTree.readDirectory(dir);

  for (unsigned i = 0; i < _files.size(); i++)
    {
      _files[i].clear();
      _eTree.getFilesForChannel(i, _files[i]);
    }


  //Check that the files all correspond to each other
  if (_files.size() > 1)
    {
      for (unsigned i = 1; i < _files.size(); i++)
	{
	  string extension;

	  if (i < _channelExts.size())
	    extension = _channelExts[i];
	  else
	    extension = _maskExt;

	  if (_files[i].size() != _files[0].size())
	    SVL_LOG(SVL_LOG_FATAL, "Channel " << i 
		    << " and channel 0 do not have the same number of frames: channel "
		    << i << " has " <<_files[i].size() << " while channel 0 has " 
		    << _files[0].size() << ". (Aborting after first difference)");
	  
	  for (unsigned j = 0; j < _files[0].size(); j++)
	    {
	      assert(!strcmp(stripExtension(_files[0][j], _channelExts[0]).c_str(), 
			     stripExtension(_files[i][j], extension).c_str()));
	    }
	}
    }
}

void svlImageLoader::readImageSeq(const char *imageSeqFilename, bool append)
{
  if (append) {
    for (unsigned ch = 0; ch < _files.size(); ch++)
      _files[ch].clear();
  }

  svlImageSequence sequence;
  sequence.load(imageSeqFilename);
  if (!hasHomogeneousExtensions(sequence)) {
    SVL_LOG(SVL_LOG_WARNING, "Sequence file " << imageSeqFilename
	    << " has multiple file extensions; be aware that they will be blindly"
	    << " stripped off and replaced by the standard channel extensions");
  }

  for (unsigned i = 0; i < sequence.size(); i++) {
    appendToFiles(sequence.directory() + string("/") + sequence[i]);
  }
}

void svlImageLoader::readImageSeq(const char *seqname, vector<string> &img_names)
{
	svlImageSequence sequence;
	sequence.load(seqname);
	if ( !hasHomogeneousExtensions(sequence) ) {
		SVL_LOG(SVL_LOG_WARNING, "Sequence file " << seqname
			<< " has multiple file extensions; be aware that they will be blindly"
			<< " stripped off and replaced by the standard channel extensions");
	}

	img_names.reserve(img_names.size() + sequence.size());
	for ( int i=0; i<int(sequence.size()); ++i )
		img_names.push_back(sequence.directory() + string("/") + sequence[i]);
}

void svlImageLoader::readCommandLine(char **args, int num_args)
{
  for (unsigned ch = 0; ch < _files.size(); ch++)
      _files[ch].clear();

  _dir = "";

  for (int index = 0; index < num_args; index++) {
    string imageFilename = string(args[index]);
    if (strExtension(imageFilename).compare("xml") == 0) {
      readImageSeq(imageFilename.c_str(), true);
    } else {
      appendToFiles(imageFilename);
    }
  }
}


void svlImageLoader::getAllFrames(vector<vector<IplImage *> > & output,
				  unsigned maxToRead)
{
  output.clear();

  bool newFrameNeeded = true;
  for (unsigned int i = 0; i < _files[0].size() && i < maxToRead; i++)
    {
      if (newFrameNeeded)
	output.push_back(vector<IplImage *>() );

      newFrameNeeded = getFrame(i, output.back());
    }
}

void svlImageLoader::getAllFrames(vector<svlMaskedFrame> & output) 
{
  output.clear();

  bool newFrameNeeded = true;
  for (unsigned int i = 0; i < _files[0].size(); i++)
    {
      if (newFrameNeeded)
	output.push_back(svlMaskedFrame() );

      newFrameNeeded = getFrame(i, output.back());
    }
}

bool svlImageLoader::getFrame(unsigned idx, vector<IplImage *> & output) 
{
  SVL_LOG(SVL_LOG_VERBOSE, "Getting frame "
	  << stripExtension(_files[0][idx], _channelExts[0]));

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    filenames.push_back(  _dir + "/" + _files[ch][idx] );

  return getFrameFromFiles(filenames, output);
}

bool svlImageLoader::getFrame(unsigned idx, svlMaskedFrame & output) 
{
  string name = stripExtension(_files[0][idx], _channelExts[0]);
  if (_useMask) {
    SVL_LOG(SVL_LOG_VERBOSE, "Getting masked frame " << name);
  } else {
    SVL_LOG(SVL_LOG_VERBOSE, "Getting frame " << name);
  }

  output.channels.clear();
  
  if (output.mask)
    cvReleaseMat(&output.mask);

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    filenames.push_back(  _dir + "/" + _files[ch][idx] );

  bool success1 = getFrameFromFiles(filenames, output.channels);

  if (success1)
    {
      if (_useMask)
	{
	  assert(output.channels.size());
	  output.mask = cvCreateMat(output.channels[0]->height,
				     output.channels[0]->width,
				     CV_32FC1);
	  assert(output.mask);

	  bool success2 = readMatrix(output.mask, (_dir + "/" + _files[numChannels()][idx]).c_str() );

	  return success2;
	}
      else
	{
	  output.mask = NULL;
	}

      return true;
    }
  else
    return false;

  //unreached
  assert(false);
  return false;
}

/*
bool svlImageLoader::getFrameFromBasename(const string & baseName,
					  vector<IplImage *> & output) 
{
  string strippedName = stripExtension(baseName);

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelExts.size(); ch++)
    filenames.push_back(strippedName + _channelExts[ch]);

  return getFrameFromFiles(filenames, output);
  }*/

IplImage *svlImageLoader::getSimpleImageFromFilename(const char *filename,
						     unsigned channel)
{
  string fullname = strWithoutExt(string(filename)) + _channelExts[channel];
  
  IplImage *img = loadFrame(fullname, channel);
  if (img == NULL) {
    SVL_LOG(SVL_LOG_MESSAGE, "...failed to load " << filename);
    return NULL;
  }

  if ((img->width != (int) _width) || (img->height != (int) _height)) {
    if (_height != -1)
      resizeInPlace(&img, _height, _width);
  }
  return img;
}

void svlImageLoader::setDepthMapSize(int width, int height)
{
  _depthMapWidth = width;
  _depthMapHeight = height;
}

void svlImageLoader::setResizeSize(int width, int height)
{

  _width = width;
  _height = height;

  if (_depthMapWidth == -1)
    setDepthMapSize(width, height);
}

unsigned svlImageLoader::numFrames() const
{
  if (_files.size())
    return (unsigned)_files[0].size();
  return 0;
}

string svlImageLoader::getFrameBasename(unsigned i) const
{
  string name;
  if (_files.size()) {
    name = stripExtension(_files[0][i], _channelExts[0]);
  }
  return name;
}

void svlImageLoader::printChannelInfo(unsigned chNum) const
{
  SVL_ASSERT(chNum < _channelPatchTypes.size());

  svlPatchDefinitionType chPatchType = _channelPatchTypes[chNum];
  string chExt = _channelExts[chNum];
  
  string msg = "Channel:\t";
  switch (chPatchType) {
  case SVL_INTENSITY_PATCH:
    msg += string("INTENSITY");
    break;
  case SVL_DEPTH_PATCH:
    msg += string("DEPTH");
    break;
  default:
    assert(false);
  }
  msg += string("\t") + chExt;
  SVL_LOG(SVL_LOG_MESSAGE, msg);
}  


// private functions ----------------------------------------------------

bool svlImageLoader::getFrameFromFiles(const vector<string> & files,
				       vector<IplImage *> & output) 
{

  if (files.size() != _channelTypes.size())
    return false;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++) {
      const string & filename = files[ch];

      IplImage * img = loadFrame(filename,  _colorTypes[ch]);

      if (img == NULL)
	{
	  SVL_LOG(SVL_LOG_MESSAGE, "...failed to load " << filename);

	  for (unsigned ch2 = 0; ch2 < ch; ch2++)
	    cvReleaseImage(&output[ch2]);

	  output.clear();

	  return false;
	}
      else
	{
	  if ((img->width != (int) _width) || (img->height != (int) _height))
	    {
	      if (_height != -1)
		{
		  resizeInPlace(&img, _height, _width);
		}
	    }

        // do the necessary color conversion for non-grayscale images
	  if (_colorTypes[ch] != SVL_COLOR_GRAY || _colorTypes[ch] != SVL_COLOR_UNDEFINED)
	    svlChangeColorModel(img, _colorTypes[ch]);

	  if (_channelTypes[ch] == SVL_CHANNEL_EDGE) {
	    IplImage *temp = cvCloneImage(_softEdgeCalculator.processImage(img));
	    cvReleaseImage(&img);
	    img = temp;
	  }
	  
	  output.push_back(img);	  
	}    
    }


  return true;
}

IplImage *svlImageLoader::loadFrame(const string & path, unsigned ch) const {
  IplImage *img = NULL;

  switch (_colorTypes[ch])
    {
    case SVL_COLOR_GRAY:
      img = cvLoadImage(path.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
      break;
    case SVL_COLOR_UNDEFINED:
      img = readMatrixAsIplImage(path.c_str(), _depthMapWidth, _depthMapHeight);
      break;
    default:
      img = cvLoadImage(path.c_str(), CV_LOAD_IMAGE_UNCHANGED);
    }
  return img;
}

void svlImageLoader::appendToFiles(const string &fileName)
{
  string strippedName = strWithoutExt(fileName);

  for (unsigned ch = 0; ch < _channelExts.size(); ch++)
    _files[ch].push_back(strippedName + _channelExts[ch]);
  
  if (_useMask)
    _files[_channelExts.size()].push_back(strippedName + _maskExt);
}


// command line processing

svlChannelType svlImageLoader::textToChannelType(const char *text) const {
  if (!strcmp(text, "INTENSITY")) 
    return SVL_CHANNEL_INTENSITY;
  
  if (!strncmp(text, "EDGE", 4)) 
    return SVL_CHANNEL_EDGE;
   
  if (!strcmp(text, "DEPTH")) 
    return SVL_CHANNEL_DEPTH; 
    
  return SVL_CHANNEL_ERROR;
}

svlColorType svlImageLoader::textToColorType(const char *text) const {
  if (!strncmp(text, "DEPTH", 5))
    return SVL_COLOR_UNDEFINED;

    // if just "EDGE", the color is gray
  if (!strncmp(text, "INTENSITY", 9) || !strcmp(text, "EDGE"))
    return SVL_COLOR_GRAY;

  // otherwise, should be EDGE_<colortype>
  if (!strncmp(text, "EDGE_", 5)) {
    const char *color = text + 5;

    if (!strcmp(color, "RGB"))
      return SVL_COLOR_BGR;
    
    if (!strcmp(color, "BG"))
      return SVL_COLOR_BG;
    
    if (!strcmp(color, "BR"))
      return SVL_COLOR_BR;
    
    if (!strcmp(color, "GR"))
      return SVL_COLOR_GR;
    
    if (!strcmp(color, "HSV"))
      return SVL_COLOR_HSV;
    
    if (!strcmp(color, "HS"))
      return SVL_COLOR_HS;
    
    if (!strcmp(color, "HV"))
      return SVL_COLOR_HV;
    
    if (!strcmp(color, "YCrCb"))
      return SVL_COLOR_YCrCb;
    
    if (!strcmp(color, "CrCb"))
      return SVL_COLOR_CrCb;
    
    if (!strcmp(text, "H"))
      return SVL_COLOR_H;
  }

  return SVL_COLOR_ERROR;
}

svlPatchDefinitionType svlImageLoader::channelTypeToPatchType(svlChannelType type) const {
  switch(type) {

  case SVL_CHANNEL_INTENSITY:
  case SVL_CHANNEL_EDGE:
    return SVL_INTENSITY_PATCH;

  case SVL_CHANNEL_DEPTH:
    return SVL_DEPTH_PATCH;

  case SVL_CHANNEL_ERROR:
    SVL_ASSERT(false); 
  }

  SVL_ASSERT(false); 
  return SVL_INTENSITY_PATCH; // to suppress warnings
}


class svlImageLoaderConfig : public svlConfigurableModule {
public:
  svlImageLoaderConfig() : svlConfigurableModule("svlVision.svlImageLoader") {}

  void readConfiguration(XMLNode& node) {
    // process everything except the channels and the extensionsToIgnore
    ((svlConfigurableModule *)this)->readConfiguration(node);

    // parse <channel type="" ext=""/>, one per channel (append to existing)
    for (int i = 0; i < node.nChildNode("channel"); i++) {
        XMLNode child = node.getChildNode("channel", i);
        const char *type = child.getAttribute("type");
        const char *extension = child.getAttribute("ext");
        SVL_ASSERT((type != NULL) && (extension != NULL));
	svlImageLoader::CHANNELS.push_back(make_pair<string, string>(type, extension));
    }

    // parse <ignore ext=""/> (append extension to existing extensions to ignore)
    for (int i = 0; i < node.nChildNode("ignore"); i++) {
        XMLNode child = node.getChildNode("ignore", i);
        const char *extension = child.getAttribute("ext");
        SVL_ASSERT(extension != NULL);
	svlImageLoader::EXT_TO_IGNORE.push_back(string(extension));
    }
  }

  void setConfiguration(const char *name, const char *value) {
    if (!strcmp(name, "defaultExtension")) {
      svlImageLoader::BASE_EXT = string(value);
    } else if (!strcmp(name, "maskExtension")) {
      svlImageLoader::MASK_EXT = string(value);
    } else if (!strcmp(name, "useMask")) {
      svlImageLoader::USE_MASK = 
	(!strcasecmp(value, "TRUE") || !strcmp(value, "1"));
    } else if (!strcmp(name, "resizeWidth")) {
      svlImageLoader::WIDTH = atoi(value);
    } else if (!strcmp(name, "resizeHeight")) {
      svlImageLoader::HEIGHT = atoi(value);
    } else if (!strcmp(name, "depthMapWidth")) {
      svlImageLoader::DEPTH_MAP_WIDTH = atoi(value);
    } else if (!strcmp(name, "depthMapHeight")) {
      svlImageLoader::DEPTH_MAP_HEIGHT = atoi(value);
    } else if (!strcmp(name, "channels")) {
      // parse channels (clear existing so command line can override config file)
      svlImageLoader::CHANNELS.clear();
      string value_copy(value);
      const char *ptr = strtok((char *)value_copy.c_str(), " ");
      while (ptr != NULL) {
 	  string channelType = string(ptr);
	  ptr = strtok(NULL, " ");
	  SVL_ASSERT_MSG(ptr, "Channel string must contain \"CHANNEL_TYPE EXTENSION\" pairs");
	  string extension = string(ptr);
	  ptr = strtok(NULL, " ");
	  svlImageLoader::CHANNELS.push_back(make_pair<string, string>(channelType, extension));
	}
    } else if (!strcmp(name, "ignoreExtensions")) {
      // parse extensions to ignore (clear existing so command line can override config file)
      svlImageLoader::EXT_TO_IGNORE.clear();
      string value_copy(value);
      const char *ptr = strtok((char *)value_copy.c_str(), " ");
      while (ptr != NULL) {
	svlImageLoader::EXT_TO_IGNORE.push_back(string(ptr));
	ptr = strtok(NULL, " ");
      }
    } else {
      SVL_LOG(SVL_LOG_FATAL, "unrecognized configuration option for " << this->name());
    }
  }
};

static svlImageLoaderConfig gImageLoaderConfig;

