/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchDefinition.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra<sidbatra@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <limits>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlPatchDefinition.h"
#include "svlVision.h"

using namespace std;

map<IplImage*,svlPatchStats*> svlIntensityPatchDefinition::patch_cache = map<IplImage*,svlPatchStats*>();

// svlPatchDefinition class --------------------------------------------------

svlPatchDefinition::svlPatchDefinition() :
  _template(NULL), _validChannel(0),  _templateSum(0), _templateNormSq(0)
{
    _validRect = cvRect(0, 0, 0, 0);
}

svlPatchDefinition::svlPatchDefinition(XMLNode& node) :
  _template(NULL), _validChannel(0), _templateSum(0), _templateNormSq(0)
{
    _validRect = cvRect(0, 0, 0, 0);
    //read(node);
}

svlPatchDefinition::svlPatchDefinition(const IplImage *t, const CvRect& rect, int channel)
{
  if (t != NULL) {
        _template = cvCloneImage(t);
    } else {
        _template = NULL;
    }
    _validRect.x = rect.x;
    _validRect.y = rect.y;
    _validRect.width = rect.width;
    _validRect.height = rect.height;
    _validChannel = channel;

    CvScalar sc = cvSum(t);
    _templateSum = sc.val[0];

    float norm = cvNorm(t);
    _templateNormSq = norm * norm;
}

svlPatchDefinition::~svlPatchDefinition()
{
    if (_template != NULL) {
        cvReleaseImage(&_template);
	
	_template = NULL;
    }
}

// constuction helper functions
svlSmartPointer<svlPatchDefinition> svlPatchDefinition::createPatchDefinition(svlPatchDefinitionType t)
{
    svlSmartPointer<svlPatchDefinition> p;
    switch (t) {
    case SVL_INTENSITY_PATCH:
        p = svlSmartPointer<svlPatchDefinition>(new svlIntensityPatchDefinition());
        break;
    case SVL_DEPTH_PATCH:
        p = svlSmartPointer<svlPatchDefinition>(new svlDepthPatchDefinition());
        break;
    default:
      break; // to suppress compiler warnings
    }

    return p;
}

svlSmartPointer<svlPatchDefinition> svlPatchDefinition::createPatchDefinition(XMLNode &node)
{
    assert(!node.isEmpty());
    svlSmartPointer<svlPatchDefinition> p;

    const char *t = node.getAttribute("type");
    assert(t != NULL);

    if (!strcasecmp(t, "intensity")) {
      p = svlSmartPointer<svlPatchDefinition>(new svlIntensityPatchDefinition(node));
    } else if (!strcasecmp(t, "depth")) { 
      p = svlSmartPointer<svlPatchDefinition>(new svlDepthPatchDefinition(node));
     } else {
        SVL_LOG(SVL_LOG_FATAL, "unknown patch type " << t);
    }

    return p;
}

bool svlPatchDefinition::read(XMLNode &node)
{		
    assert(!node.isEmpty());
    if (_template != NULL) {
        cvReleaseImage(&_template);
    }

    svlPatch *tmp = readPatch(node);
    _validChannel = tmp->channel;
    _validRect = tmp->rect;
    _template = cvCloneImage(tmp->patch);

    CvScalar sc = cvSum(_template);
    _templateSum = sc.val[0];

    float norm = cvNorm(_template);
    _templateNormSq = norm * norm;
    delete tmp;

    return true;
}

// Read a patch and return as a struct pointer.
svlPatch* svlPatchDefinition::readPatch(XMLNode &node)
{
	svlPatch *ret = new svlPatch();

	// Read valid region
	if ( node.getAttribute("validRect") ) {
		vector<int> v;
		parseString<int>(node.getAttribute("validRect"), v);
		assert(v.size() == 4);
		ret->rect = cvRect(v[0], v[1], v[2], v[3]);
	} else {
		ret->rect = cvRect(0, 0, 0, 0);
	}

	// Read channel
	if ( node.getAttribute("channel") ) {
		ret->channel = atoi(node.getAttribute("channel"));
	} else {
		ret->channel = 0;
	}

	// Read patch size
	int w = 0;
	int h = 0;
	if (node.getAttribute("size")) {
		vector<int> v;
		parseString<int>(node.getAttribute("size"), v);
		assert(v.size() == 2);
		w = v[0]; h = v[1];
	}

	// Read patch data
	if ( w > 0 && h > 0 ) {
		ret->patch = cvCreateImage(cvSize(w, h), IPL_DEPTH_32F, 1);
		vector<float> v;
		parseString<float>(node.getText(), v);
		assert(v.size() == (unsigned)(w * h));
		unsigned n = 0;
		for (unsigned y = 0; y < (unsigned)ret->patch->height; y++) {
			float *p = &CV_IMAGE_ELEM(ret->patch, float, y, 0);
			for (unsigned x = 0; x < (unsigned)ret->patch->width; x++) {
				p[x] = v[n++];
			}
		}
	}

	return ret;
}

bool svlPatchDefinition::write(ostream& os)
{
    assert(_template != NULL);
    SVL_ASSERT_MSG(_template->depth == IPL_DEPTH_32F, "All patches now expected to have 32F depth");

    os << "  <PatchDefinition type=\"";
    switch (this->patchType()) 
	{
    case SVL_INTENSITY_PATCH:
        os << "intensity";
        break;
    case SVL_DEPTH_PATCH:
        os << "depth";
        break;
    default:
        SVL_LOG(SVL_LOG_FATAL, "unknown patch type");
    }
    os << "\"\n"
       << "    validRect=\"" << _validRect.x << " " << _validRect.y << " "
       << _validRect.width << " " << _validRect.height << "\"\n"
       << "    channel=\"" << _validChannel << "\"\n"
       << "    size=\"" << _template->width << " " << _template->height << "\">\n";
    
    for (unsigned y = 0; y < (unsigned)_template->height; y++) 
	{
        os << "   ";
        for (unsigned x = 0; x < (unsigned)_template->width; x++) 
		{
                os << " " << CV_IMAGE_ELEM(_template, float, y, x);
        }        
        os << "\n";
    }
    
    os << "  </PatchDefinition>\n";

    return true;
}

static double correlationResponse(IplImage *img, IplImage *tmpl) {
  IplImage *response = cvCreateImage(cvSize(img->width - tmpl->width + 1,
					    img->height - tmpl->height + 1),
				     IPL_DEPTH_32F, 1);
  // do not subtract the mean of each patch, just do
  // normalized cross-correlation (so expansion doesn't
  // affect it negatively)
  cvMatchTemplate(img, tmpl, response, CV_TM_CCORR_NORMED);  

  double min_value, max_value;
  cvMinMaxLoc(response, &min_value, &max_value);

  cvReleaseImage(&response);
  return max_value;
}

static double correlationResponseWithExpansion(IplImage *imBiggerWidth, IplImage *imBiggerHeight) {

  int w_small = imBiggerHeight->width;
  int w_large = imBiggerWidth->width;

  int h_small = imBiggerWidth->height;
  int h_large = imBiggerHeight->height;

  // first, expand in height
  IplImage *expanded = cvCreateImage(cvSize(w_large, h_large * 2 - h_small),
				     imBiggerWidth->depth, imBiggerWidth->nChannels);

  cvZero(expanded);
  cvSetImageROI(expanded, cvRect(0, h_large - h_small, w_large, h_small));
  cvCopy(imBiggerWidth, expanded);
  cvResetImageROI(expanded);

  double val1 = correlationResponse(expanded, imBiggerHeight);
  cvReleaseImage(&expanded);

  // now expand in width
  expanded = cvCreateImage(cvSize(w_large * 2 - w_small, h_large),
			   imBiggerHeight->depth, imBiggerHeight->nChannels);
  cvZero(expanded);
  cvSetImageROI(expanded, cvRect(w_large - w_small, 0, w_small, h_large));
  cvCopy(imBiggerHeight, expanded);
  cvResetImageROI(expanded);

  double val2 = correlationResponse(expanded, imBiggerWidth);
  cvReleaseImage(&expanded);

  //cerr << " (" << val1 << ", " << val2 << ")";

  return max(val1, val2);
}

static bool enoughOverlap(CvRect rect1, int w1, int h1, CvRect rect2, int w2, int h2) {
  // coordinates of the overlap area
  int x1 = max(rect1.x, rect2.x);
  int y1 = max(rect1.y, rect2.y);
  int x2 = min(rect1.x + rect1.width + w1, rect2.x + rect2.width + w2);
  int y2 = min(rect1.y + rect1.height + h1, rect2.y + rect2.height + h2);

  return ((x2 - x1) >= min(w1, w2)) && ((y2-y1) >= min(h1, h2));
}

bool svlPatchDefinition::isCorrelatedWith(svlPatchDefinition *patch, double threshold) const {
  // checks if they correspond to the same region within the image; if they don't,
  // then they're definitely not correlated

  // assumes that both are valid entries, and thus are non-uniform
  // also assumes that they're of the same type

  if (!isSameType(patch)) {
    //cerr << "not same type" << endl;
    return false;
  }

  if (_template->nChannels != patch->_template->nChannels) {
    return false;
  }

  int a_width = _template->width;
  int b_width = patch->_template->width;

  int a_height = _template->height;
  int b_height = patch->_template->height;

  //cerr << a_width << "x" << a_height << ", " << b_width << "x" << b_height << ": ";

  // if the templates come from completely different areas of the image,
  // they're obviously not correlated
  // ("completely different" is defined as even if the two patches
  // are positioned with as much overlap as possible within the valid
  // rectangle, they still overlap by less than min(w1, w2) x min(h1, h2))
  if (!enoughOverlap(_validRect, a_width, a_height,
		     patch->_validRect, b_width, b_height)) {
    //cerr << "not enough overlap" << endl;
    return false;
  }

  double value = 0;

  if (a_width <= b_width && a_height <= b_height) {
    // patch->_template strictly bigger
    //cerr << "patch strictly bigger, ";
    value = correlationResponse(patch->_template, _template);
  
  } else if (b_width <= a_width && b_height <= a_height) {
    // template strictly bigger
    //cerr << "templ strictly bigger, ";
    value = correlationResponse(_template, patch->_template);
  
  } else if (a_width < b_width) {
    // patch is bigger in width, but template is bigger in height
    //cerr << "patch bigger in w, ";
    value = correlationResponseWithExpansion(patch->_template, _template);
  
  } else if (b_width < a_width) {
    // template is strictly bigger in width
    //cerr << "patch bigger in h, ";
    value = correlationResponseWithExpansion(_template, patch->_template);
  
  } else {
    assert(false); // debugging to make sure all cases are considered
  }

  //cerr << value << endl;
  return (value > threshold);
}

IplImage *svlPatchDefinition::responseImage(const IplImage *image) const
{
  SVL_ASSERT_MSG(image->nChannels == 1, "responseImage not implemented yet for multiple channels");

  static int handle = svlCodeProfiler::getHandle("svlPatchDefinition::responseImage");
  svlCodeProfiler::tic(handle);

  assert((_template != NULL));
  assert((image != NULL));
  
  int width = image->width - _template->width +1;
  int height = image->height - _template->height+1;
  
  IplImage *response = cvCreateImage(cvSize(width, height), IPL_DEPTH_32F, 1);
  cvMatchTemplate(image, _template, response, CV_TM_CCORR);
  
  svlCodeProfiler::toc(handle);
  return response;
}

IplImage *svlPatchDefinition::responseImage(const std::vector<IplImage *>& images) const
{
    assert(images.size() > (unsigned)_validChannel);
    return responseImage(images[_validChannel]);
}


// Combines all the logic into one private function; either the responseImage and
// integral images are precomputed and the optimization is easy, or they aren't
// and the original image is passed in instead so everything happens on the spot
double svlPatchDefinition::patchValueHelper(const IplImage *image,
					    const CvPoint& location, float normConstant,
					    const IplImage *responseImage,
					    const IplImage *imageSum, const IplImage *imageSumSq) const {

  bool optimized = responseImage && imageSum && imageSumSq;
  SVL_ASSERT(optimized || image);

  if (!optimized)
    SVL_ASSERT(image->depth == IPL_DEPTH_32F);

  SVL_ASSERT(_template->depth == IPL_DEPTH_32F);

  int startX = location.x + _validRect.x;
  int startY = location.y + _validRect.y;

  int N = _template->width * _template->height;
  float normConstantSq = normConstant * normConstant;
  float wTemplateSum = normConstant * _templateSum;

  const double *ps = NULL;
  const double *ps2 = NULL;
  const double *psSq = NULL;
  const double *psSq2 = NULL;
  const float *psR = NULL;
  if (optimized) {
    ps = &CV_IMAGE_ELEM(imageSum, double, startY, 0);
    ps2 = &CV_IMAGE_ELEM(imageSum, double, startY + _template->height, 0);
    psSq = &CV_IMAGE_ELEM(imageSumSq, double, startY, 0);
    psSq2 = &CV_IMAGE_ELEM(imageSumSq, double, startY + _template->height, 0);
    psR = &CV_IMAGE_ELEM(responseImage, float, startY, 0);
  }

  float windowSum = 0;
  float windowSumSq = 0;
  float ccorr = 0;
  
  // iterate over the _validRect region
  float maxValSq = -numeric_limits<float>::max();
  for (int y = startY; y < startY + _validRect.height; y++) {
    for (int x = startX; x < startX + _validRect.width; x++) {
      if (optimized) {
	ccorr = psR[x];
	windowSum = (float)(ps[x] + ps2[x + _template->width] - 
			    ps[x + _template->width] - ps2[x]);
	windowSumSq = (float)(psSq[x] + psSq2[x + _template->width] - 
			      psSq[x + _template->width] - psSq2[x]);
      } 
      
      // not optimized, everything has to be computed from scratch
      else {
	const float *pw = &CV_IMAGE_ELEM(image, float, y, x);
	const float *pp = (const float *)_template->imageData;
	ccorr = windowSum = windowSumSq = 0.0;
	for (int v = 0; v < _template->height; v++) {
	  for (int u = 0; u < _template->width; u++) {
	    windowSum += pw[u];
	    windowSumSq += pw[u] * pw[u];
	    ccorr += pp[u] * pw[u];
	  }
	  pp += _template->width;
	  pw += image->width;
	}
      }

      // the sum over the template: sqrt(\sum_{x',y'} (I(x+x', y+y') - normConstant)^2)
      float windowNormSq = windowSumSq - 2 * normConstant * windowSum + N * normConstantSq;

      // the cross-correlation of _template with (I - normConstant)
      // (wTemplateSum = weightedTemplateSum = _templateSum * normConstant)
      float normedCcorr = (ccorr - wTemplateSum);

      // for efficiency, will work with the square of everything
      float normedCcorrSq = (normedCcorr > 0) ? 
	(normedCcorr * normedCcorr) : 
	(-1 * normedCcorr * normedCcorr);

      // ... normalized 
      // (at the end of the loop will also be normalized by _templateNormSq)
      normedCcorrSq /= windowNormSq;

      if (normedCcorrSq > maxValSq)
	maxValSq = normedCcorrSq;

      if (isinf(normedCcorr) || isnan(normedCcorr) || isnan(ccorr) || isinf(ccorr) ||
	  isnan(windowNormSq)) { // || abs(maxValSq) > _templateNormSq) {
	cerr << "windowSum: " << windowSum << ", windowSumSq: " << windowSumSq << ", normConstant: " << normConstant << ", N: " << N << endl;
	cerr << "ccorr: " << ccorr << ", windowNormSq: " << windowNormSq << ", normedCcorr: " << normedCcorr << ", _templateNormSq: " << _templateNormSq << endl;
	cerr << "_templateSum: " << _templateSum << endl;
	if (optimized) {
	  cerr << "ps values: " << ps[x] << " " << ps[x+_template->width] << " " << ps2[x] << " " << ps2[x+_template->width] << endl;
	  cerr << "ps2 values: " << psSq[x] << " " << psSq[x+_template->width] << " " << psSq2[x] << " " << psSq2[x+_template->width] << endl;
	} else {
	  cerr << "Full patch (32x32):" << endl;
	  for (int j = 0; j < 32; j++) {
	    cerr << "\t";
	    for (int i = 0; i < 32; i++) {
	      cerr << CV_IMAGE_ELEM(image, float, location.y+j, location.x+i) << " ";
	    }
	    cerr << endl;
     	  }

	  cerr << "Patch:" << endl;
	  for (int j = 0; j < _template->height; j++) {
	    cerr << "\t";
	    for (int i = 0; i < _template->width; i++) {
	      cerr << CV_IMAGE_ELEM(image, float, y+j, x+i) << " ";
	    }
	    cerr << endl;
	  }
	}
	cerr << "Template:" << endl;
	for (int j = 0; j < _template->height; j++) {
	  cerr << "\t";
	  for (int i = 0; i < _template->width; i++) {
	    cerr << CV_IMAGE_ELEM(_template, float, j, i) << " ";
	  }
	  cerr << endl;
	}
	cerr << "Cross-correlation:" << endl;
	const float *pw = &CV_IMAGE_ELEM(image, float, y, x);
	const float *pp = (const float *)_template->imageData;
	for (int v = 0; v < _template->height; v++) {
	  cerr << "\t";
	  for (int u = 0; u < _template->width; u++) {
	    cerr << pp[u] * pw[u] << " ";
	  }
	  pp += _template->width;
	  pw += image->width;
	  cerr << endl;
	}
      } 

    } // end x iter
    
    if (optimized) {
      ps += imageSum->width;
      ps2 += imageSum->width;
      psSq += imageSumSq->width;
      psSq2 += imageSumSq->width;
      psR += responseImage->width;
    }

  } // end y iter

  maxValSq /= _templateNormSq;

  // take care to zero out any invalid values (will occur e.g. if the patch is uniform)
  if (isnan(maxValSq) || isinf(maxValSq)) return 0;
  return (maxValSq > 0) ? sqrt(maxValSq) : -1 * sqrt(-1 * maxValSq);
}

double svlPatchDefinition::patchValue(const CvPoint& location, float normConstant,
				      const IplImage *responseImage,
				      const IplImage *imageSum, const IplImage *imageSumSq) const 
{
  return patchValueHelper(NULL, location, normConstant, responseImage, imageSum, imageSumSq);
}

double svlPatchDefinition::patchValue(const IplImage *image, const CvPoint& location, float normConstant) const
{
  return patchValueHelper(image, location, normConstant, NULL, NULL, NULL);
}

vector<double> svlPatchDefinition::patchValues(const vector<IplImage *>& images,
					       const vector<CvPoint>& locations,
					       const vector<float> &normConstants) const
{
    assert(images.size() > (unsigned)_validChannel);
    return patchValues(images[_validChannel], locations, normConstants);
}

vector<double> svlPatchDefinition::patchValues(const IplImage *image,
					       const vector<CvPoint>& locations,
					       const vector<float>& normConstants) const
{
  vector<double> values;
  values.reserve(locations.size());
  
  for (unsigned i = 0; i < locations.size(); i++) {
    values.push_back(patchValueHelper(image, locations[i], normConstants[i], NULL, NULL, NULL));
  }

  return values;
}









// svlIntensityPatchDefinition class --------------------------------------------------

svlIntensityPatchDefinition::svlIntensityPatchDefinition() :
    svlPatchDefinition()
{
    // do nothing
}

svlIntensityPatchDefinition::svlIntensityPatchDefinition(XMLNode& node) :
  svlPatchDefinition(node)
{
    // do nothing
	read(node);
}

svlIntensityPatchDefinition::svlIntensityPatchDefinition(const IplImage *t, const CvRect& rect, int channel) :
  svlPatchDefinition(t, rect, channel)
{
    // do nothing

}

svlIntensityPatchDefinition::~svlIntensityPatchDefinition()
{
    // do nothing
}










// svlDepthPatchDefinition class --------------------------------------------------

// initialize static data members
int svlDepthPatchDefinition::_maxDistance = 3;
int svlDepthPatchDefinition::_totalBins = 100;
CvSize svlDepthPatchDefinition::_windowSize;

svlDepthPatchDefinition::svlDepthPatchDefinition() :
  svlPatchDefinition()
{
    // do nothing
}

svlDepthPatchDefinition::svlDepthPatchDefinition(XMLNode& node) :
  svlPatchDefinition(node)
{
    // do nothing
	read(node);
}

svlDepthPatchDefinition::svlDepthPatchDefinition(const IplImage *t, const CvRect& rect, int channel) :
  svlPatchDefinition(t, rect, channel)
{
    // do nothing
}

svlDepthPatchDefinition::~svlDepthPatchDefinition()
{
    // do nothing
}



//Smooth histogram with a kernel of [1/3 1/3 1/3]
void svlDepthPatchDefinition::smoothHistogram(int *histogram, int totalLength)
{
   
	//Apply smoothing kernel
   for (int i = 1; i < totalLength - 1 ; i++)
   {
		histogram[i] = (histogram[i-1] + histogram[i] + histogram[i+1] ) / 3;  
   }

   //Handle first and last entry seperately
   histogram[0] = ( histogram[0] + histogram[1] ) / 2;
   histogram[totalLength-1] = ( histogram[totalLength-2] + histogram[totalLength-1] ) / 2;
}





//Computes the earth movers distances between the two given integer arrays (histograms)
int svlDepthPatchDefinition::computeEMD(int *a , int *b , int totalLength)
{
	float *af = new float[totalLength];
	float *bf = new float[totalLength];

	//Convert histograms to float types to use opencv emd function
	for( int i=0 ; i<totalLength ; i++)
	{
		af[i] = (float)a[i];
		bf[i] = (float)b[i];
	}

	float dist = cvCalcEMD(af,totalLength,bf,totalLength,0,CV_DIST_L1);

	delete []af;
	delete []bf;

	return (int)dist;
}


//Computes the euclidean distances between the two given integer arrays (histograms)
int svlDepthPatchDefinition::computeEuclideanDistance(int *a , int *b , int totalLength)
{

	float totalDistance = 0;

	for( int i=0 ; i<totalLength ; i++)
	{
            totalDistance += (a[i] - b[i]) * (a[i] - b[i]);
	}

	return (int)sqrt(totalDistance);
}













